__all__ = [
    'ClassifierModelStatsComposite',
    'RegressionModelStatsComposite',
    'IndividualPredictionsComposite',
    'ShapDependenceComposite',
    'ShapInteractionsComposite',
    'DecisionTreesComposite',
]

import dash_bootstrap_components as dbc
import dash_html_components as html

from.dashboard_methods import *
from .classifier_components import *
from .regression_components import *
from .overview_components import *
from .connectors import *
from .shap_components import *
from .decisiontree_components import *


class ClassifierModelStatsComposite(ExplainerComponent):
    def __init__(self, explainer, title="Classification Stats", name=None,
                    hide_title=False, hide_selector=True, pos_label=None,
                    bin_size=0.1, quantiles=10, cutoff=0.5):
        """Composite of multiple classifier related components: 
            - precision graph
            - confusion matrix
            - lift curve
            - classification graph
            - roc auc graph
            - pr auc graph

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Decision Trees".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_title (bool, optional): hide title. Defaults to False.          
            hide_selector (bool, optional): hide all pos label selectors. Defaults to True.
            pos_label ({int, str}, optional): initial pos label. Defaults to explainer.pos_label
            bin_size (float, optional): bin_size for precision plot. Defaults to 0.1.
            quantiles (int, optional): number of quantiles for precision plot. Defaults to 10.
            cutoff (float, optional): initial cutoff. Defaults to 0.5.
        """
        super().__init__(explainer, title, name)

        self.hide_title = hide_title

        self.precision = PrecisionComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)
        self.confusionmatrix = ConfusionMatrixComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)
        self.liftcurve = LiftCurveComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)
        self.classification = ClassificationComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)
        self.rocauc = RocAucComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)
        self.prauc = PrAucComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)

        self.cutoffpercentile = CutoffPercentileComponent(explainer, hide_selector=hide_selector, pos_label=pos_label)
        self.cutoffconnector = CutoffConnector(self.cutoffpercentile,
                [self.precision, self.confusionmatrix, self.liftcurve, 
                 self.classification, self.rocauc, self.prauc])

        self.register_components(
            self.precision, self.confusionmatrix, self.liftcurve,
            self.classification, self.rocauc, self.prauc, self.cutoffpercentile,
            self.cutoffconnector)

    def layout(self):
        return html.Div([
            dbc.Row([
                make_hideable(
                    dbc.Col([
                     html.H2('Model Performance:')]), hide=self.hide_title),
            ]),
            dbc.Row([
                dbc.Col([
                    self.cutoffpercentile.layout(),
                ])
            ]),
            dbc.Row([
                dbc.Col([
                    self.precision.layout()
                ], md=6, align="start"),
                dbc.Col([
                    self.confusionmatrix.layout()
                ], md=6, align="start"),              
            ]),
            dbc.Row([
                dbc.Col([
                    self.liftcurve.layout()         
                ], md=6, align="start"),
                dbc.Col([
                    self.classification.layout()
                ], md=6, align="start"),
            ]),
            dbc.Row([    
                dbc.Col([
                    self.rocauc.layout()
                ], md=6),
                dbc.Col([
                    self.prauc.layout()
                ], md=6),
            ]),
        ])


class RegressionModelStatsComposite(ExplainerComponent):
    def __init__(self, explainer, title="Regression Stats", name=None,
                    hide_title=False,
                    logs=False, pred_or_actual="vs_pred", residuals='difference',
                    col=None):
        """Composite for displaying multiple regression related graphs:

        - predictions vs actual plot
        - residual plot
        - residuals vs feature

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Regression Stats".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_title (bool, optional): hide title. Defaults to False.
            logs (bool, optional): Use log axis. Defaults to False.
            pred_or_actual (str, optional): plot residuals vs predictions 
                        or vs y (actual). Defaults to "vs_pred".
            residuals (str, {'difference', 'ratio', 'log-ratio'} optional): 
                    How to calcualte residuals. Defaults to 'difference'.
            col ({str, int}, optional): Feature to use for residuals plot. Defaults to None.
        """
        super().__init__(explainer, title, name)
     
        self.hide_title = hide_title
        assert pred_or_actual in ['vs_actual', 'vs_pred'], \
            "pred_or_actual should be 'vs_actual' or 'vs_pred'!"

        self.preds_vs_actual = PredictedVsActualComponent(explainer, logs=logs)
        self.modelsummary = RegressionModelSummaryComponent(explainer)
        self.residuals = ResidualsComponent(explainer, 
                            pred_or_actual=pred_or_actual, residuals=residuals)
        self.residuals_vs_col = ResidualsVsColComponent(explainer, 
                                    col=col, residuals=residuals)
        self.register_components([self.preds_vs_actual, self.modelsummary,
                    self.residuals, self.residuals_vs_col])

    def layout(self):
        return html.Div([
            dbc.Row([
                make_hideable(
                    dbc.Col([
                        html.H2('Model Performance:')]), hide=self.hide_title)
            ]),
            dbc.Row([
                dbc.Col([
                    self.preds_vs_actual.layout()
                ], md=6),
                dbc.Col([
                    self.modelsummary.layout()     
                ], md=6),
            ], align="start"),
            dbc.Row([
                dbc.Col([
                    self.residuals.layout()
                ], md=6),
                dbc.Col([
                    self.residuals_vs_col.layout()
                ], md=6),
            ])
        ])


class IndividualPredictionsComposite(ExplainerComponent):
    def __init__(self, explainer, title="Individual Predictions", name=None,
                        hide_title=False, hide_selector=True):
        """Composite for a number of component that deal with individual predictions:

        - random index selector
        - prediction summary
        - shap contributions graph
        - shap contribution table
        - pdp graph

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Individual Predictions".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_title (bool, optional): hide title. Defaults to False.
            hide_selector(bool, optional): hide all pos label selectors. Defaults to True.
        """
        super().__init__(explainer, title, name)

        self.hide_title = hide_title

        if self.explainer.is_classifier:
            self.index = ClassifierRandomIndexComponent(
                            explainer, hide_selector=hide_selector)
        elif self.explainer.is_regression:
            self.index = RegressionRandomIndexComponent(explainer)
        self.summary = PredictionSummaryComponent(
                            explainer, hide_selector=hide_selector)
        self.contributions = ShapContributionsGraphComponent(
                                explainer, hide_selector=hide_selector)
        self.pdp = PdpComponent(
                        explainer, hide_selector=hide_selector)
        self.contributions_list = ShapContributionsTableComponent(
                                        explainer, hide_selector=hide_selector)

        self.index_connector = IndexConnector(self.index, 
                [self.summary, self.contributions, self.pdp, self.contributions_list])

        self.register_components(
            self.index, self.summary, self.contributions, 
            self.pdp, self.contributions_list, self.index_connector)

    def layout(self):
        return html.Div([
                dbc.Row([
                    dbc.Col([
                        self.index.layout()
                    ]),
                    dbc.Col([
                        self.summary.layout()
                    ])
                ]),
                dbc.Row([
                    dbc.Col([
                        self.contributions.layout()
                    ]),
                    dbc.Col([
                        self.pdp.layout()
                    ]),
                ]),
                dbc.Row([
                    dbc.Col([
                        self.contributions_list.layout()
                    ]),
                    dbc.Col([
                        html.Div([]),
                    ]),
                ])
        ])


class ShapDependenceComposite(ExplainerComponent):
    def __init__(self, explainer, title='Feature Dependence', name=None,
                    hide_selector=True,
                    depth=None, cats=True):
        """Composite of ShapSummary and ShapDependence component

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Feature Dependence".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_selector (bool, optional): hide all pos label selectors. Defaults to True.
            depth (int, optional): Number of features to display. Defaults to None.
            cats (bool, optional): Group categorical features. Defaults to True.
        """
        super().__init__(explainer, title, name)
        
        self.shap_summary = ShapSummaryComponent(
                    self.explainer, 
                    hide_selector=hide_selector, 
                    depth=depth, cats=cats)
        self.shap_dependence = ShapDependenceComponent(
                    self.explainer, 
                    hide_selector=hide_selector, hide_cats=True, 
                    cats=cats)
        self.connector = ShapSummaryDependenceConnector(
                    self.shap_summary, self.shap_dependence)
        self.register_components(self.shap_summary, self.shap_dependence, self.connector)

    def layout(self):
        return dbc.Container([
            dbc.Row([
                dbc.Col([
                    self.shap_summary.layout()
                ], md=6),
                dbc.Col([
                    self.shap_dependence.layout()
                ], md=6),
                ]),
            ],  fluid=True)


class ShapInteractionsComposite(ExplainerComponent):
    def __init__(self, explainer, title='Feature Interactions', name=None,
                    hide_selector=True,
                    depth=None, cats=True):
        """Composite of InteractionSummaryComponent and InteractionDependenceComponent

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Feature Interactions".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_selector (bool, optional): hide all pos label selectors. Defaults to True.
            depth (int, optional): Initial number of features to display. Defaults to None.
            cats (bool, optional): Initally group cats. Defaults to True.
        """
        super().__init__(explainer, title, name)

        self.interaction_summary = InteractionSummaryComponent(
                explainer, hide_selector=hide_selector, depth=depth, cats=cats)
        self.interaction_dependence = InteractionDependenceComponent(
                explainer, hide_selector=hide_selector, cats=cats, hide_cats=True)
        self.connector = InteractionSummaryDependenceConnector(
            self.interaction_summary, self.interaction_dependence)
        self.register_components(
            self.interaction_summary, self.interaction_dependence, self.connector)
        
    def layout(self):
        return html.Div([
                dbc.Row([
                    dbc.Col([
                        self.interaction_summary.layout()
                    ], width=6),
                    dbc.Col([
                        self.interaction_dependence.layout()
                    ], width=6),
                ]) 
        ])


class DecisionTreesComposite(ExplainerComponent):
    def __init__(self, explainer, title="Decision Trees", name=None,
                    hide_selector=True):
        """Composite of decision tree related components:
        
        - individual decision trees barchart
        - decision path table
        - deciion path graph

        Args:
            explainer (Explainer): explainer object constructed with either
                        RandomForestClassifierExplainer() or RandomForestRegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Decision Trees".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_selector (bool, optional): hide all pos label selectors. Defaults to True.
        """
        super().__init__(explainer, title, name)
        
        self.trees = DecisionTreesComponent(explainer, hide_selector=hide_selector)
        self.decisionpath_table = DecisionPathTableComponent(explainer, hide_selector=hide_selector)

        if explainer.is_classifier:
            self.index = ClassifierRandomIndexComponent(explainer, hide_selector=hide_selector)
        elif explainer.is_regression:
            self.index = RegressionRandomIndexComponent(explainer)

        self.decisionpath_graph = DecisionPathGraphComponent(explainer, hide_selector=hide_selector)

        self.index_connector = IndexConnector(self.index, 
            [self.trees, self.decisionpath_table, self.decisionpath_graph])
        self.highlight_connector = HighlightConnector(self.trees, 
            [self.decisionpath_table, self.decisionpath_graph])

        self.register_components(self.index, self.trees, 
                self.decisionpath_table, self.decisionpath_graph, 
                self.index_connector, self.highlight_connector)

        # elif explainer.is_regression:
        #     self.index = RegressionRandomIndexComponent(explainer)
        #     self.decisionpath_graph = DecisionPathGraphComponent(explainer, hide_selector=hide_selector)# DummyComponent()
        #     self.index_connector = IndexConnector(self.index, 
        #             [self.trees, self.decisionpath_table])
        #     self.highlight_connector = HighlightConnector(
        #             self.trees, [self.decisionpath_table, self.decisionpath_graph])

        #     self.register_components(
        #             self.index, self.trees, self.decisionpath_table,
        #             self.decisionpath_graph, self.index_connector, 
        #             self.highlight_connector)
        
    def layout(self):
        return html.Div([
            dbc.Row([
                dbc.Col([
                    self.index.layout(),
                ])
            ]),
            dbc.Row([
                dbc.Col([
                    self.trees.layout(),
                ])
            ]),
            dbc.Row([
                dbc.Col([
                    self.decisionpath_table.layout()
                ])
            ]),
            dbc.Row([
                dbc.Col([
                    self.decisionpath_graph.layout()
                ])
            ]),
        ])



__all__ = [
    'PredictedVsActualComponent',
    'ResidualsComponent',
    'ResidualsVsColComponent',
    'RegressionModelSummaryComponent',
]

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from .dashboard_methods import *


class PredictedVsActualComponent(ExplainerComponent):
    def __init__(self, explainer, title="Predicted vs Actual", name=None,
                    hide_title=False, hide_log_x=False, hide_log_y=False,
                    logs=False, log_x=False, log_y=False):
        """Shows a plot of predictions vs y.

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Predicted vs Actual".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_title (bool, optional) Hide the title. Defaults to False.
            hide_log_x (bool, optional): Hide the log_x toggle. Defaults to False.
            hide_log_y (bool, optional): Hide the log_y toggle. Defaults to False.
            logs (bool, optional): Whether to use log axis. Defaults to False.
            log_x (bool, optional): log only x axis. Defaults to False.
            log_y (bool, optional): log only y axis. Defaults to False.
        """
        super().__init__(explainer, title, name)
        self.hide_title, self.hide_log_x, self.hide_log_y = \
            hide_title, hide_log_x, hide_log_y
        
        self.logs, self.log_x, self.log_y = logs, log_x, log_y
        self.register_dependencies(['preds'])

    def layout(self):
        return html.Div([
            dbc.Row([
                make_hideable(html.H3("Predictions"), hide=self.hide_title)
            ]),
            dbc.Row([
                make_hideable(
                    dbc.Col([
                        dbc.FormGroup(
                        [
                            # html.Label("Log y"),
                            dbc.RadioButton(
                                id='pred-vs-actual-logy-'+self.name,
                                className="form-check-input",
                                checked=self.log_y),   
                            dbc.Label("Log y",
                                    html_for='pred-vs-actual-logy-'+self.name,
                                    className="form-check-label"), 
                        ], check=True),
                    ], md=1, align="center"), hide=self.hide_log_y),
                dbc.Col([
                    dcc.Loading(id="loading-pred-vs-actual-graph-"+self.name, 
                                children=[dcc.Graph(id='pred-vs-actual-graph-'+self.name)]),
                ], md=11)
            ]),
            dbc.Row([
                make_hideable(
                    dbc.Col([
                        dbc.FormGroup(
                        [
                            dbc.RadioButton(
                                id='pred-vs-actual-logx-'+self.name,
                                className="form-check-input",
                                checked=self.log_x),
                            dbc.Label("Log x",
                                    html_for='pred-vs-actual-logx-'+self.name,
                                    className="form-check-label"),   
                        ], check=True),
                    ], md=2), hide=self.hide_log_x),
            ], justify="center")   
        ])

    def _register_callbacks(self, app):
        @app.callback(
            Output('pred-vs-actual-graph-'+self.name, 'figure'),
            [Input('pred-vs-actual-logx-'+self.name, 'checked'),
             Input('pred-vs-actual-logy-'+self.name, 'checked')],
        )
        def update_predicted_vs_actual_graph(log_x, log_y):
            return self.explainer.plot_predicted_vs_actual(log_x=log_x, log_y=log_y)

class ResidualsComponent(ExplainerComponent):
    def __init__(self, explainer, title="Residuals", name=None,
                    hide_title=False, hide_pred_or_actual=False, hide_ratio=False,
                    pred_or_actual="vs_pred", residuals="difference"):
        """Residuals plot component

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Residuals".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_title (bool, optional) Hide the title. Defaults to False.
            hide_pred_or_actual (bool, optional): hide vs predictions or vs 
                        actual for x-axis toggle. Defaults to False.
            hide_ratio (bool, optional): hide residual type dropdown. Defaults to False.
            pred_or_actual (str, {'vs_actual', 'vs_pred'}, optional): Whether 
                        to plot actual or predictions on the x-axis. 
                        Defaults to "vs_pred".
            residuals (str, {'difference', 'ratio', 'log-ratio'} optional): 
                    How to calcualte residuals. Defaults to 'difference'.
        """
        super().__init__(explainer, title, name)

        self.hide_title, self.hide_pred_or_actual = hide_title, hide_pred_or_actual
        self.hide_ratio = hide_ratio
        self.pred_or_actual = pred_or_actual
        self.residuals = residuals
        assert residuals in ['difference', 'ratio', 'log-ratio'], \
            ("parameter residuals should in ['difference', 'ratio', 'log-ratio']"
             f" but you passed residuals={residuals}")
        self.register_dependencies(['preds', 'residuals'])

    def layout(self):
        return html.Div([
            dbc.Row([
                make_hideable(html.H3("Residuals"), hide=self.hide_title)
            ]),
            dbc.Row([
                dbc.Col([
                    dcc.Loading(id="loading-residuals-graph-"+self.name, 
                                children=[dcc.Graph(id='residuals-graph-'+self.name)]),
                ])
            ]),
            dbc.Row([
                make_hideable(
                    dbc.Col([
                        dbc.FormGroup(
                        [
                            dbc.RadioItems(
                                options=[
                                    {"label": "vs Prediction", "value": "vs_pred"},
                                    {"label": "vs Actual", "value": "vs_actual"},
                                ],
                                value=self.pred_or_actual,
                                id='residuals-pred-or-actual-'+self.name,
                                inline=True,
                            ),
                        ]),
                    ], md=3), hide=self.hide_pred_or_actual),
                make_hideable(
                    dbc.Col([
                        html.Label('Residual type:'),
                        dcc.Dropdown(id='residuals-type-'+self.name,
                                options = [{'label': 'Difference', 'value': 'difference'},
                                            {'label': 'Ratio', 'value': 'ratio'},
                                            {'label': 'Log ratio', 'value': 'log-ratio'}],
                                value=self.residuals),
                    ], md=3), hide=self.hide_ratio),
            ], justify="center")  
        ])

    def register_callbacks(self, app):
        @app.callback(
            Output('residuals-graph-'+self.name, 'figure'),
            [Input('residuals-pred-or-actual-'+self.name, 'value'),
             Input('residuals-type-'+self.name, 'value')],
        )
        def update_residuals_graph(pred_or_actual, residuals):
            vs_actual = pred_or_actual=='vs_actual'
            return self.explainer.plot_residuals(vs_actual=vs_actual, residuals=residuals)

class ResidualsVsColComponent(ExplainerComponent):
    def __init__(self, explainer, title="Residuals vs feature", name=None,
                    hide_title=False, hide_col=False, hide_ratio=False, hide_cats=False, 
                    hide_points=False, hide_winsor=False,
                    col=None, residuals='difference', cats=True, 
                    points=True, winsor=0):
        """Show residuals vs a particular Feature component

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Residuals vs feature".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
            hide_title (bool, optional) Hide the title. Defaults to False.
            hide_col (bool, optional): Hide de column selector. Defaults to False.
            hide_ratio (bool, optional): Hide the  toggle. Defaults to False.
            hide_cats (bool, optional): Hide group cats toggle. Defaults to False.
            hide_points (bool, optional): Hide group points toggle. Defaults to False.
            hide_winsor (bool, optional): Hide winsor input. Defaults to False.
            col ([type], optional): Initial feature to display. Defaults to None.
            residuals (str, {'difference', 'ratio', 'log-ratio'} optional): 
                    How to calcualte residuals. Defaults to 'difference'.
            cats (bool, optional): group categorical columns. Defaults to True.
            points (bool, optional): display point cloud next to violin plot 
                    for categorical cols. Defaults to True
            winsor (int, 0-50, optional): percentage of outliers to winsor out of 
                    the y-axis. Defaults to 0.
        """
        super().__init__(explainer, title, name)
        self.hide_title = hide_title
        self.hide_col, self.hide_ratio, self.hide_cats, self.hide_points, self.hide_winsor = \
            hide_col, hide_ratio, hide_cats, hide_points, hide_winsor
        self.col, self.residuals, self.cats, self.points, self.winsor = \
            col, residuals, cats, points, winsor
        if self.col is None:
            self.col = self.explainer.columns_ranked_by_shap(self.cats)[0]
        
        assert residuals in ['difference', 'ratio', 'log-ratio'], \
            ("parameter residuals should in ['difference', 'ratio', 'log-ratio']"
             f" but you passed residuals={residuals}")
        self.register_dependencies(['preds', 'residuals'])

    def layout(self):
        return html.Div([
            dbc.Row([
                make_hideable(html.H3("Residuals vs Feature"), hide=self.hide_title)
            ]),
            dbc.Row([
                make_hideable(
                    dbc.Col([
                        dbc.Label("Column:"),
                        dcc.Dropdown(id='residuals-vs-col-col-'+self.name,
                            options=[{'label': col, 'value':col} 
                                            for col in self.explainer.columns_ranked_by_shap(self.cats)],
                            value=self.col),
                    ], md=4), hide=self.hide_col),
                make_hideable(
                        dbc.Col([
                            dbc.Label("Grouping:"),
                            dbc.FormGroup(
                            [
                                dbc.RadioButton(
                                    id='residuals-vs-col-group-cats-'+self.name, 
                                    className="form-check-input",
                                    checked=self.cats),
                                dbc.Label("Group Cats",
                                        html_for='residuals-vs-col-group-cats-'+self.name,
                                        className="form-check-label"),
                            ], check=True),
                        ], md=2), self.hide_cats),
            ]),
            dbc.Row([
                dcc.Loading(id="loading-residuals-vs-col-graph-"+self.name, 
                                children=[dcc.Graph(id='residuals-vs-col-graph-'+self.name)]),
            ]),
            dbc.Row([
                make_hideable(
                        dbc.Col([
                            html.Label('Residual type:'),
                            dcc.Dropdown(id='residuals-vs-col-residuals-type-'+self.name,
                                    options = [{'label': 'Difference', 'value': 'difference'},
                                                {'label': 'Ratio', 'value': 'ratio'},
                                                {'label': 'Log ratio', 'value': 'log-ratio'}],
                                    value=self.residuals),
                        ], md=3), hide=self.hide_ratio),
                make_hideable(
                        dbc.Col([ 
                            dbc.Label("Winsor:"),
                            dbc.Input(id='residuals-vs-col-winsor-'+self.name, 
                                    value=self.winsor,
                                type="number", min=0, max=49, step=1),
                        ], md=2), hide=self.hide_winsor),  
                make_hideable(
                        dbc.Col([
                            dbc.Label("Points:"),
                            dbc.FormGroup(
                            [
                                dbc.RadioButton(
                                    id='residuals-vs-col-show-points-'+self.name, 
                                    className="form-check-input",
                                    checked=self.points),
                                dbc.Label("Show points",
                                        html_for='residuals-vs-col-show-points-'+self.name,
                                        className="form-check-label"),
                            ], check=True),
                        ],  md=3), self.hide_points),
            ]),
        ])

    def register_callbacks(self, app):
        @app.callback(
            Output('residuals-vs-col-graph-'+self.name, 'figure'),
            [Input('residuals-vs-col-col-'+self.name, 'value'),
             Input('residuals-vs-col-residuals-type-'+self.name, 'value'),
             Input('residuals-vs-col-show-points-'+self.name, 'checked'),
             Input('residuals-vs-col-winsor-'+self.name, 'value')],
        )
        def update_residuals_graph(col, residuals, points, winsor):
            return self.explainer.plot_residuals_vs_feature(
                        col, residuals=residuals, points=points, 
                        winsor=winsor, dropna=True)

        @app.callback(
            Output('residuals-vs-col-col-'+self.name, 'options'),
            [Input('residuals-vs-col-group-cats-'+self.name, 'checked')])
        def update_dependence_shap_scatter_graph(cats):
            return [{'label': col, 'value': col} 
                for col in self.explainer.columns_ranked_by_shap(cats)]

class RegressionModelSummaryComponent(ExplainerComponent):
    def __init__(self, explainer, title="Model Summary", name=None):
        """Show model summary statistics (RMSE, MAE, R2) component

        Args:
            explainer (Explainer): explainer object constructed with either
                        ClassifierExplainer() or RegressionExplainer()
            title (str, optional): Title of tab or page. Defaults to 
                        "Model Summary".
            name (str, optional): unique name to add to Component elements. 
                        If None then random uuid is generated to make sure 
                        it's unique. Defaults to None.
        """
        super().__init__(explainer, title, name)
        self.register_dependencies(['preds', 'residuals'])

    def layout(self):
        return html.Div([
            dcc.Markdown(id='model-summary-'+self.name, children=self.explainer.metrics_markdown())
        ])
